import re

BASE64URL_REGEX = r"^([a-z0-9_-]{4})*($|[a-z0-9_-]{3}$|[a-z0-9_-]{2}$)$"


def is_valid_jwt(value: str) -> bool:
    """Checks if value seems to be a JWT without attempting to decode it."""
    if not isinstance(value, str):
        return False

    value = value.strip()

    # Valid JWT must have exactly 2 periods (Header.Payload.Signature)
    if value.count(".") != 2:
        return False

    # Each of the parts must be base64 encoded
    for part in value.split("."):
        if not part or not re.search(BASE64URL_REGEX, part, re.IGNORECASE):
            return False

    return True



import asyncio
import time
from typing import cast, Dict, Optional

import httpx
from httpx import AsyncClient as AsyncHttpClient

from tws.base.client import TWSClient, ClientException


class AsyncClient(TWSClient):
    """Asynchronous client implementation for TWS API interactions.

    Provides asynchronous methods for interfacing with the TWS API.
    """

    def __init__(self, public_key: str, secret_key: str, api_url: str):
        """Initialize the asynchronous client.

        Args:
            public_key: The TWS public key
            secret_key: The TWS secret key
            api_url: The base URL for your TWS API instance
        """
        super().__init__(public_key, secret_key, api_url)
        self.session = cast(AsyncHttpClient, self.session)

    def create_session(
        self,
        base_url: str,
        headers: Dict[str, str],
    ) -> AsyncHttpClient:
        """Create a new asynchronous HTTP session.

        Args:
            base_url: The base URL for the API
            headers: Dictionary of HTTP headers to include in requests

        Returns:
            A configured asynchronous HTTPX client instance
        """
        return AsyncHttpClient(
            base_url=base_url,
            headers=headers,
            follow_redirects=True,
            http2=True,
        )

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        # Close the underlying HTTP session
        await self.session.aclose()

    async def _make_request(
        self,
        method: str,
        uri: str,
        payload: Optional[dict] = None,
        params: Optional[dict] = None,
    ):
        """Make a HTTP request to the TWS API.

        Args:
            method: HTTP method to use (GET, POST, etc)
            uri: API endpoint URI
            payload: Optional request body data
            params: Optional URL query parameters

        Returns:
            Parsed JSON response from the API

        Raises:
            ClientException: If a request error occurs
        """
        try:
            response = await self.session.request(
                method, f"/rest/v1/{uri}", json=payload, params=params
            )
            response.raise_for_status()
            return response.json()
        except httpx.RequestError as e:
            raise ClientException(f"Request error occurred: {e}")

    async def _make_rpc_request(
        self, function_name: str, payload: Optional[dict] = None
    ):
        """Make an RPC request to the TWS API.

        Args:
            function_name: Name of the RPC function to call
            payload: Optional request body data

        Returns:
            Parsed JSON response from the API
        """
        return await self._make_request("POST", f"rpc/{function_name}", payload)

    async def run_workflow(
        self,
        workflow_definition_id: str,
        workflow_args: dict,
        timeout=600,
        retry_delay=1,
        tags: Optional[Dict[str, str]] = None,
    ):
        self._validate_workflow_params(timeout, retry_delay)
        self._validate_tags(tags)

        payload = {
            "workflow_definition_id": workflow_definition_id,
            "request_body": workflow_args,
        }
        if tags is not None:
            payload["tags"] = tags

        try:
            result = await self._make_rpc_request("start_workflow", payload)
        except httpx.HTTPStatusError as e:
            if (
                e.response.status_code == 400
                and e.response.json().get("code") == "P0001"
            ):
                raise ClientException("Workflow definition ID not found")
            raise ClientException(f"HTTP error occurred: {e}")

        workflow_instance_id = result["workflow_instance_id"]
        start_time = time.time()

        while True:
            self._check_timeout(start_time, timeout)

            params = {"select": "status,result", "id": f"eq.{workflow_instance_id}"}
            result = await self._make_request(
                "GET", "workflow_instances", params=params
            )

            if not result:
                raise ClientException(
                    f"Workflow instance {workflow_instance_id} not found"
                )

            instance = result[0]
            workflow_result = self._handle_workflow_status(instance)
            if workflow_result is not None:
                return workflow_result

            await asyncio.sleep(retry_delay)



from .base.client import ClientException

from ._sync.client import SyncClient as Client

from ._async.client import AsyncClient

__all__ = [
    "AsyncClient",
    "Client",
    "ClientException",
]



from abc import ABC, abstractmethod
import re
import time
from typing import Optional, Union, Coroutine, Any, Dict
from urllib.parse import urlparse

from httpx import Client as SyncClient, AsyncClient

from tws.utils import is_valid_jwt


class ClientException(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class TWSClient(ABC):
    def __init__(
        self,
        public_key: str,
        secret_key: str,
        api_url: str,
    ):
        if not public_key:
            raise ClientException("Public key is required")
        if not secret_key:
            raise ClientException("Secret key is required")
        if not api_url:
            raise ClientException("API URL is required")

        # Secret key must be a valid UUID v4
        if not re.match(
            r"^[0-9a-f]{8}-[0-9a-f]{4}-4[0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}$",
            secret_key,
            re.IGNORECASE,
        ):
            raise ClientException("Malformed secret key")

        # API URL must be a valid URL
        if urlparse(api_url).scheme not in {"https", "http"}:
            raise ClientException("Malformed API URL")

        # Public key should look like a valid JWT
        if not is_valid_jwt(public_key):
            raise ClientException("Malformed public key")

        base_url = api_url.rstrip("/")
        headers = {
            "Authorization": secret_key,
            "apikey": public_key,
            "Content-Type": "application/json",
        }
        self.session = self.create_session(base_url, headers)

    @abstractmethod
    def create_session(
        self,
        base_url: str,
        headers: Dict[str, str],
    ) -> Union[SyncClient, AsyncClient]:
        raise NotImplementedError()

    @staticmethod
    def _validate_workflow_params(
        timeout: Union[int, float],
        retry_delay: Union[int, float],
    ) -> None:
        if not isinstance(timeout, (int, float)) or timeout < 1 or timeout > 3600:
            raise ClientException("Timeout must be between 1 and 3600 seconds")
        if (
            not isinstance(retry_delay, (int, float))
            or retry_delay < 1
            or retry_delay > 60
        ):
            raise ClientException("Retry delay must be between 1 and 60 seconds")

    @staticmethod
    def _handle_workflow_status(instance: dict) -> Optional[dict]:
        status = instance.get("status")

        # TODO also handle CANCELLED state
        if status == "COMPLETED":
            return instance.get("result", {})
        elif status == "FAILED":
            raise ClientException(
                f"Workflow execution failed: {instance.get('result', {})}"
            )
        return None

    @staticmethod
    def _check_timeout(start_time: float, timeout: Union[int, float]) -> None:
        if time.time() - start_time > timeout:
            raise ClientException(
                f"Workflow execution timed out after {timeout} seconds"
            )

    @staticmethod
    def _validate_tags(tags: Optional[Dict[str, str]]) -> None:
        if tags is not None:
            if not isinstance(tags, dict):
                raise ClientException("Tags must be a dictionary")
            for key, value in tags.items():
                if not isinstance(key, str) or not isinstance(value, str):
                    raise ClientException("Tag keys and values must be strings")
                if len(key) > 255 or len(value) > 255:
                    raise ClientException(
                        "Tag keys and values must be <= 255 characters"
                    )

    @abstractmethod
    def run_workflow(
        self,
        workflow_definition_id: str,
        workflow_args: dict,
        timeout=600,
        retry_delay=1,
        tags: Optional[Dict[str, str]] = None,
    ) -> Union[dict, Coroutine[Any, Any, dict]]:
        """Execute a workflow and wait for it to complete or fail.

        Args:
            workflow_definition_id: The unique identifier of the workflow definition to execute
            workflow_args: Dictionary of arguments to pass to the workflow
            timeout: Maximum time in seconds to wait for workflow completion (1-3600)
            retry_delay: Time in seconds between status checks (1-60)
            tags: Optional dictionary of tag key-value pairs to attach to the workflow

        Returns:
            The workflow execution result as a dictionary

        Raises:
            ClientException: If the workflow fails, times out, or if invalid parameters are provided
        """
        pass
